"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const fs = require("fs");
const querystring = require("querystring");
let routes = new Map();
routes.set('GET', []);
routes.set('POST', []);
routes.set('PUT', []);
routes.set('DELETE', []);
routes.set('PATCH', []);
routes.set('HEAD', []);
let config = {
    ignorecase: undefined,
    root: undefined,
    logger: undefined,
    onInit(method, path) {
        if (config.logger && config.logger.isInfoEnabled())
            config.logger.info(`${method.toUpperCase()} ${path}`);
    },
    handlers: undefined
};
const Type = {
    RegExp: 1,
    RegExpMaped: 2,
    String: 0
};
function RouteConfig(controllerDirs, opts) {
    if (!opts)
        throw new Error('Could not found route config');
    if (!opts.onInit)
        opts.onInit = config.onInit;
    if (opts.handlers) {
        opts.handlers = opts.handlers.filter(e => e);
        if (opts.handlers.length === 0)
            opts.handlers = undefined;
    }
    config = opts;
    if (typeof controllerDirs === 'string')
        controllerDirs = [controllerDirs];
    const loadClass = (dir, classes = []) => {
        const stat = fs.statSync(dir);
        if (stat.isFile()) {
            if (/\.js$/.test(dir)) {
                if (!classes.includes(dir))
                    classes.push(dir);
            }
        }
        else if (stat.isDirectory()) {
            const files = fs.readdirSync(dir);
            for (const f of files) {
                classes = classes.concat(loadClass(path.join(dir, f), classes));
            }
        }
        return classes;
    };
    controllerDirs.forEach(e => {
        loadClass(path.resolve(e)).forEach(cls => require(cls));
    });
    return async (ctx, next) => {
        if (ctx.routed)
            return await next();
        const r = routes.get(ctx.req.method);
        if (!r)
            ctx.throws(405, `Method ${ctx.req.method} not supported`);
        for (const route of r) {
            if (route.isValid(ctx)) {
                if (config.logger && config.logger.isDebugEnabled())
                    config.logger.debug('<--', `${route.method} ${route.rawPath}`);
                ctx.routed = true;
                if (config.handlers) {
                    for (const handler of config.handlers) {
                        await handler.onRouted(ctx);
                    }
                }
                await route.executeMiddle(ctx);
                return await next();
            }
        }
        ctx.throws(404, 'Could not found API');
    };
}
exports.RouteConfig = RouteConfig;
var route;
(function (route) {
    function bindRoutes(method, path, ...cbs) {
        routes.get(method).push(new RouteTemplate(method, path, cbs));
    }
    route.bindRoutes = bindRoutes;
})(route = exports.route || (exports.route = {}));
function INJECT(cb) {
    return function (_target, _fieldName, func) {
        if (!func.value['mid'])
            func.value['mid'] = new Array();
        func.value['mid'].splice(0, 0, cb);
    };
}
exports.INJECT = INJECT;
function ALL(pathStr) {
    return _api('ALL', pathStr);
}
exports.ALL = ALL;
function GET(pathStr) {
    return _api('GET', pathStr);
}
exports.GET = GET;
function POST(pathStr) {
    return _api('POST', pathStr);
}
exports.POST = POST;
function PUT(pathStr) {
    return _api('PUT', pathStr);
}
exports.PUT = PUT;
function DELETE(pathStr) {
    return _api('DELETE', pathStr);
}
exports.DELETE = DELETE;
function HEAD(pathStr) {
    return _api('HEAD', pathStr);
}
exports.HEAD = HEAD;
function PATCH(pathStr) {
    return _api('PATCH', pathStr);
}
exports.PATCH = PATCH;
function _api(method, pathStr) {
    if (typeof pathStr === 'string' && config.root) {
        pathStr = `${config.root}${!pathStr.endsWith('/') ? pathStr : pathStr.substr(0, pathStr.length - 1)}`;
    }
    return function (_target, _fieldName, func) {
        if (!func.value['meta'])
            func.value['meta'] = {};
        func.value['meta'].route = {
            method,
            path: pathStr
        };
        const mid = (func.value['mid'] || []);
        mid.push(async (ctx, next) => {
            ctx.data = await func.value(ctx, next);
            if (config.handlers) {
                for (const handler of config.handlers) {
                    await handler.onFinished(ctx);
                }
            }
        });
        if (method !== 'ALL') {
            route.bindRoutes(method, pathStr, ...mid);
        }
        else {
            for (let k in routes) {
                route.bindRoutes(k, pathStr, ...mid);
            }
        }
        config.onInit(method, pathStr);
    };
}
class RouteTemplate {
    constructor(method, rawPath, cbs) {
        this.method = method;
        this.rawPath = rawPath;
        this.cbs = cbs;
        this.type = Type.String;
        if (typeof rawPath === 'string') {
            if (rawPath.includes(':')) {
                let key = rawPath.replace(/:[^\/]+/g, '([^\\/]+)');
                let [, ...m] = rawPath.match(new RegExp(`${key}`));
                this.path = new RegExp(`^${key}$`, config.ignorecase ? 'i' : undefined);
                this.paramsName = m.map(e => e.substr(1));
                this.type = Type.RegExpMaped;
            }
            else if (config.ignorecase) {
                this.path = rawPath.toLowerCase();
            }
            else {
                this.path = rawPath;
            }
        }
        else if (rawPath instanceof RegExp) {
            this.path = rawPath;
            this.type = Type.RegExp;
        }
        else {
            throw new Error('Error route path');
        }
    }
    isValid(ctx) {
        let [pathname, query] = ctx.req.url.split('?');
        if (this.type === Type.String) {
            if (config.ignorecase) {
                if (this.path !== pathname.toLowerCase()) {
                    return false;
                }
            }
            else if (this.path !== pathname) {
                return false;
            }
        }
        else if (this.type === Type.RegExpMaped) {
            const matchRoute = pathname.match(this.path);
            if (!matchRoute)
                return false;
            ctx.params = this.paramsName.reduce((sum, e, i) => {
                sum[e] = decodeURIComponent(matchRoute[i + 1]);
                return sum;
            }, {});
        }
        else {
            const matchRoute = pathname.match(this.path);
            if (!matchRoute)
                return false;
            ctx.params = matchRoute.reduce((sum, e, i) => {
                if (i > 0)
                    sum.push(decodeURIComponent(e));
                return sum;
            }, []);
        }
        ctx.query = query ? querystring.parse(query) : {};
        return true;
    }
    async executeMiddle(ctx) {
        let i = 0;
        const cbs = this.cbs;
        const len = cbs.length;
        let next = async (n = 1) => {
            i += n;
            if (i < len)
                await cbs[i](ctx, next);
        };
        await cbs[i](ctx, next);
    }
}
