import { Context, Server } from 'hinos'
import * as path from 'path'
import * as fs from 'fs'
import * as querystring from 'querystring'
import { Logger } from 'log4js'

let routes = new Map() as Map<String, RouteTemplate[]>
routes.set('GET', [])
routes.set('POST', [])
routes.set('PUT', [])
routes.set('DELETE', [])
routes.set('PATCH', [])
routes.set('HEAD', [])

let config = {
  ignorecase: undefined as boolean,
  root: undefined as string,
  logger: undefined as Logger,
  onInit(method: string, path: string | RegExp) {
    if (config.logger && config.logger.isInfoEnabled()) config.logger.info(`${method.toUpperCase()} ${path}`)
  },
  handlers: undefined as {
    onRouted: (_ctx: Context) => void | Promise<void>,
    onFinished: (_ctx: Context) => void | Promise<void>
  }[]
}

const Type = {
  RegExp: 1,
  RegExpMaped: 2,
  String: 0
}

export function RouteConfig(
  controllerDirs: string | string[],
  opts: {
    logger?: Logger,
    ignorecase?: boolean,
    root?: string,
    onInit?: (method: string, path: string | RegExp) => void,
    handlers?: {
      onRouted: (_ctx: Context) => void | Promise<void>,
      onFinished: (_ctx: Context) => void | Promise<void>
    }[]
  }): Server.Middleware {
  if (!opts) throw new Error('Could not found route config')
  if (!opts.onInit) opts.onInit = config.onInit
  if (opts.handlers) {
    opts.handlers = opts.handlers.filter(e => e)
    if (opts.handlers.length === 0) opts.handlers = undefined
  }
  config = opts as any

  if (typeof controllerDirs === 'string') controllerDirs = [controllerDirs]
  const loadClass = (dir: string, classes = []) => {
    const stat = fs.statSync(dir)
    if (stat.isFile()) {
      if (/\.js$/.test(dir)) {
        if (!classes.includes(dir)) classes.push(dir)
      }
    } else if (stat.isDirectory()) {
      const files = fs.readdirSync(dir)
      for (const f of files) {
        classes = classes.concat(loadClass(path.join(dir, f), classes))
      }
    }
    return classes
  }
  controllerDirs.forEach(e => {
    loadClass(path.resolve(e)).forEach(cls => require(cls))
  })
  // Server.Middleware to handle each request

  return async (ctx: Context, next: Function) => {
    if (ctx.routed) return await next()
    const r = routes.get(ctx.req.method)
    if (!r) ctx.throws(405, `Method ${ctx.req.method} not supported`)
    for (const route of r) {
      if (route.isValid(ctx)) {
        if (config.logger && config.logger.isDebugEnabled()) config.logger.debug('<--', `${route.method} ${route.rawPath}`)
        ctx.routed = true
        if (config.handlers) {
          for (const handler of config.handlers) {
            await handler.onRouted(ctx)
          }
        }
        await route.executeMiddle(ctx)
        return await next()
      }
    }
    ctx.throws(404, 'Could not found API')
  }
}

export namespace route {
  export function bindRoutes(method: string, path: string | RegExp, ...cbs: Server.Middleware[]) {
    routes.get(method).push(new RouteTemplate(method, path, cbs))
  }
}
export function INJECT(cb: Server.Middleware): Function {
  return function (_target: Function, _fieldName: String, func: TypedPropertyDescriptor<Function>) {
    if (!func.value['mid']) func.value['mid'] = new Array<Server.Middleware>()
    func.value['mid'].splice(0, 0, cb)
  }
}
export function ALL(pathStr: string | RegExp): Function {
  return _api('ALL', pathStr)
}
export function GET(pathStr: string | RegExp): Function {
  return _api('GET', pathStr)
}
export function POST(pathStr: string | RegExp): Function {
  return _api('POST', pathStr)
}
export function PUT(pathStr: string | RegExp): Function {
  return _api('PUT', pathStr)
}
export function DELETE(pathStr: string | RegExp): Function {
  return _api('DELETE', pathStr)
}
export function HEAD(pathStr: string | RegExp): Function {
  return _api('HEAD', pathStr)
}
export function PATCH(pathStr: string | RegExp): Function {
  return _api('PATCH', pathStr)
}

function _api(method: string, pathStr: string | RegExp) {
  if (typeof pathStr === 'string' && config.root) {
    pathStr = `${config.root}${!pathStr.endsWith('/') ? pathStr : pathStr.substr(0, pathStr.length - 1)}`
  }
  return function (_target: Function, _fieldName: String, func: TypedPropertyDescriptor<Function>) {
    if (!func.value['meta']) func.value['meta'] = {}
    func.value['meta'].route = {
      method,
      path: pathStr
    }
    const mid = (func.value['mid'] || []) as Server.Middleware[]
    mid.push(async (ctx: Context, next: Function) => {
      ctx.data = await func.value(ctx, next)
      if (config.handlers) {
        for (const handler of config.handlers) {
          await handler.onFinished(ctx)
        }
      }
    })
    if (method !== 'ALL') {
      route.bindRoutes(method, pathStr, ...mid)
    } else {
      for (let k in routes) {
        route.bindRoutes(k, pathStr, ...mid)
      }
    }
    config.onInit(method, pathStr)
  }
}

class RouteTemplate {

  private paramsName: string[]
  private type = Type.String
  private path: RegExp | string

  constructor(public method, public rawPath: string | RegExp, private cbs: Function[]) {
    if (typeof rawPath === 'string') {
      if (rawPath.includes(':')) {
        let key = rawPath.replace(/:[^\/]+/g, '([^\\/]+)')
        let [, ...m] = rawPath.match(new RegExp(`${key}`))
        this.path = new RegExp(`^${key}$`, config.ignorecase ? 'i' : undefined)
        this.paramsName = m.map(e => e.substr(1))
        this.type = Type.RegExpMaped
      } else if (config.ignorecase) {
        this.path = rawPath.toLowerCase()
      } else {
        this.path = rawPath
      }
    } else if (rawPath instanceof RegExp) {
      this.path = rawPath
      this.type = Type.RegExp
    } else {
      throw new Error('Error route path')
    }
  }

  isValid(ctx: Context) {
    let [pathname, query] = ctx.req.url.split('?')
    if (this.type === Type.String) {
      if (config.ignorecase) {
        if (this.path !== pathname.toLowerCase()) {
          return false
        }
      } else if (this.path !== pathname) {
        return false
      }
    } else if (this.type === Type.RegExpMaped) {
      const matchRoute = pathname.match(this.path)
      if (!matchRoute) return false
      ctx.params = this.paramsName.reduce((sum, e, i) => {
        sum[e] = decodeURIComponent(matchRoute[i + 1])
        return sum
      }, {})
    } else {
      const matchRoute = pathname.match(this.path)
      if (!matchRoute) return false
      ctx.params = matchRoute.reduce((sum, e, i) => {
        if (i > 0) sum.push(decodeURIComponent(e))
        return sum
      }, [])
    }
    ctx.query = query ? querystring.parse(query) : {}
    return true
  }

  async executeMiddle(ctx: Context) {
    let i = 0
    const cbs = this.cbs
    const len = cbs.length
    let next = async (n = 1) => {
      i += n
      if (i < len) await cbs[i](ctx, next)
    }
    await cbs[i](ctx, next)
  }
}
